require "defines"
require "migration"
local Under_Debug=0

if Under_Debug then
  require "debug_tool"
end  

local SLIPSTREAM_VERSION = "1.1.0"
local sschests = {}
local energy_per_action = 5000

local next = next
local north = defines.direction.north
local east = defines.direction.east
local south = defines.direction.south
local west = defines.direction.west

game.on_init(function()
	onLoad()
end)

game.on_load(function()
	onLoad()
end)

game.on_save(function()
	global.slipstream.sschests = sschests
end)

game.on_event(defines.events.on_built_entity, function(event)
	if string.find(event.created_entity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.created_entity)
	end
end)

game.on_event(defines.events.on_robot_built_entity, function(event)
	if string.find(event.created_entity.name, "^slipstream%-chest%-[%a%-]+$") then
		slipstreamBuilt(event.created_entity)
	end
end)

game.on_event(defines.events.on_tick, function(event)
	if string.match(game.tick,'^%d+(0)$') then
		for i,v in ipairs(sschests) do
			if v.entity.valid then
				chestCollection(v)
--				if v.output_config.limit > 0 then 
				  chestSpawn(v) 
--				 end
			else
				entityRemoved()
			end
		end
	end
end)

function onLoad()
	if not global.slipstream then
		global.slipstream = {}
		global.slipstream.sschests = {}
		global.slipstream.version = SLIPSTREAM_VERSION
	end
	
	if global.slipstream.version ~= SLIPSTREAM_VERSION then
		slipstreamMigration(global.slipstream.version, SLIPSTREAM_VERSION)
	end
	
	sschests = global.slipstream.sschests
end

function entityRemoved()
	for i,v in ipairs(sschests) do
		if not v.entity.valid then
			v.battery.destroy()
			table.remove(sschests,i)
		end
	end
end

-- Construct the data table for each chest as it's built.
function slipstreamBuilt(entity)
	local chest_name = entity.name
	local surface    = entity.surface
	
	local new_chest = {}
	new_chest.output_config = {}
	
	-- Record position of chest and search area.
	local pos = {x = entity.position.x, y = entity.position.y}
	local search = {{pos.x - 0.5,pos.y - 0.5},{pos.x + 0.5, pos.y + 0.5}}
	
	entity.destroy() -- Remove initial chest.
	
	-- Creates the accumulator entity that serves as the chest's power supply.
	surface.create_entity{name="slipstream-chest-battery", position = pos, force = game.forces.player}
	
	-- Recreate the chest on top of the accumulator.
	surface.create_entity{name = chest_name, position = pos, force = game.forces.player}
	
	-- Find the entity references for the battery/chest.
	local t = surface.find_entities_filtered{area = search, name = "slipstream-chest-battery"}
	new_chest.battery = t[1]
	new_chest.battery.minable = false
	new_chest.battery.destructible = false
	local t = surface.find_entities_filtered{area = search, name = chest_name}
	new_chest.entity = t[1]
	new_chest.inventory = t[1].get_inventory(defines.inventory.chest)
	
--	-- Add coordinates for collection/spawning of items.
--	new_chest.coords_collection = configureChestCollection(pos)
--	
--	-- Autoconfigure output based on dummy's direction and presence of transport belt.
--	new_chest.coords_spawn, new_chest.coords_spawn_catchment, new_chest.output_config.limit = configureChestSpawn(pos)

	new_chest.output_config.target = 1
	new_chest.output_config.limit   =1
	
	table.insert(sschests, new_chest)
end

-- Configure the area that the chest will search for new items.
function configureChestCollection(pos)
	return {
		{pos.x - 0.75,pos.y - 0.75},
		{pos.x + 0.75,pos.y + 0.75}
	}
end

-- Configure the eight locations the chest will attempt to place items down on (in all cardinal directions) if configured to do so.
function configureChestSpawn(pos)
	local spawn_coords = { -- Set up output coordinates.
		[0] = {x = pos.x - 0.25, y = pos.y - 0.64},
		[1] = {x = pos.x + 0.25, y = pos.y - 0.64},
		[2] = {x = pos.x + 0.64, y = pos.y - 0.25},
		[3] = {x = pos.x + 0.64, y = pos.y + 0.25},
		[4] = {x = pos.x + 0.25, y = pos.y + 0.64},
		[5] = {x = pos.x - 0.25, y = pos.y + 0.64},
		[6] = {x = pos.x - 0.75, y = pos.y + 0.25},
		[7] = {x = pos.x - 0.75, y = pos.y - 0.25}
	}
	
	local beltscan_coords = { -- Points to search for transport belts.
		[north] = {{pos.x - 0.1, pos.y - 1.1},{pos.x + 0.1, pos.y - 0.9}},
		[east] = {{pos.x + 0.9, pos.y - 0.1},{pos.x + 1.1, pos.y + 0.1}},
		[south] = {{pos.x - 0.1, pos.y + 0.9},{pos.x + 0.1, pos.y + 1.1}},
		[west] = {{pos.x - 1.1, pos.y - 0.1},{pos.x - 0.9, pos.y + 0.1}}
	}
	
	local directions = {north, east, south, west} -- For the for loop.
	local facing_directions = {[north] = south, [east] = west, [south] = north, [west] = east} -- If a found belt is facing in this direction, ignore it.
	local spawns_list = {}
	local limit = 0
	
--	local out_str="" 
	for i,v in ipairs(directions) do
		local belt = game.get_surface(1).find_entities_filtered{area = beltscan_coords[v], type = "transport-belt"} -- Search for a transport belt.
		if belt[1] ~= nil and belt[1].direction ~= facing_directions[v] then -- If belt is found, add two spawn locations to coords_spawn and +2 to output limit counter.
			table.insert(spawns_list,spawn_coords[v])
			table.insert(spawns_list,spawn_coords[v+1])
--      table.insert(spawns_list,{beltscan_coords[v],facing_directions[v]})
			limit = limit + 2
		end
	end
	
--	game.makefile("slipstream/dbg.txt", out_str)
	 
	-- Creates a pair of small search squares around where output items are deposited. These are searched to determine if anything is in them, since canplaceentity seems bugged.
	local spawns_catchment = {}
	for i,v in ipairs(spawns_list) do
		spawns_catchment[i] = {{v.x - 0.2, v.y - 0.2},{v.x + 0.2, v.y + 0.2}}
	end
	
	-- TODO: Add check and configuration for belt_to_ground.
	
	return spawns_list, spawns_catchment, limit
end

function chestCollection(chest) -- Pick up stuff fed towards the square the chest is placed in.
-- todo: round-robin for pickup
  local pos = {x = chest.entity.position.x, y = chest.entity.position.y}
  local beltscan_coords = { -- Points to search for transport belts.
    [north] = {{pos.x - 0.1, pos.y - 1.1},{pos.x + 0.1, pos.y - 0.9}},
    [east] = {{pos.x + 0.9, pos.y - 0.1},{pos.x + 1.1, pos.y + 0.1}},
    [south] = {{pos.x - 0.1, pos.y + 0.9},{pos.x + 0.1, pos.y + 1.1}},
    [west] = {{pos.x - 1.1, pos.y - 0.1},{pos.x - 0.9, pos.y + 0.1}}
  }
  
  local directions = {north, east, south, west} -- For the for loop.
  local facing_directions = {[north] = south, [east] = west, [south] = north, [west] = east} -- If a found belt is facing in this direction, ignore it.

  for i,v in ipairs(directions) do
    local belt = chest.entity.surface.find_entities_filtered{area = beltscan_coords[v], type = "transport-belt"} -- Search for a transport belt.
    if belt[1] ~= nil  and belt[1].direction == facing_directions[v] then --  If belt is found, add two spawn locations to coords_spawn and +2 to output limit counter.
--      debugLog("ssc " .. tostring(v) .." - ".. tostring(belt[1].name).." - ".. tostring(belt[1].type) .." - ".. tostring(belt[1].direction).." : " ..tostring(facing_directions[v]))
      for laneI=1,2,1 do
        if chest.battery.energy < energy_per_action then
          return 
        end
      
        local line=belt[1].get_transport_line(laneI)
        local cnt=line.get_item_count()
        local cmap=line.get_contents()
        local tmp=""
        if cnt ~= 0 then
          for res_name,res_count in pairs(cmap) do
            local r_stack={name = res_name, count = math.min(math.floor(chest.battery.energy/energy_per_action),res_count)}
            if r_stack["count"] > 0 and chest.inventory.can_insert(r_stack) then
              chest.inventory.insert(r_stack)
              line.remove_item (r_stack)
              chest.battery.energy = chest.battery.energy - r_stack["count"]*energy_per_action
            else
              return
            end
          end
        end
      end
    end
   end	
end

-- Place items in inventory on spawn positions.
function chestSpawn(chest)
-- todo: round-robin for spawn  
  local pos = {x = chest.entity.position.x, y = chest.entity.position.y}
  local beltscan_coords = { -- Points to search for transport belts.
    [north] = {{pos.x - 0.1, pos.y - 1.1},{pos.x + 0.1, pos.y - 0.9}},
    [east] = {{pos.x + 0.9, pos.y - 0.1},{pos.x + 1.1, pos.y + 0.1}},
    [south] = {{pos.x - 0.1, pos.y + 0.9},{pos.x + 0.1, pos.y + 1.1}},
    [west] = {{pos.x - 1.1, pos.y - 0.1},{pos.x - 0.9, pos.y + 0.1}}
  }
  
  local directions = {north, east, south, west} -- For the for loop.
  local facing_directions = {[north] = north, [east] = east, [south] = south, [west] = west} 
  
  local max_spawn=8
  for i,v in ipairs(directions) do
    local belt = chest.entity.surface.find_entities_filtered{area = beltscan_coords[v], type = "transport-belt"} -- Search for a transport belt.

    if belt[1] ~= nil and belt[1].direction == facing_directions[v]   then --  If belt is found, add two spawn locations to coords_spawn and +2 to output limit counter.
      for laneI=1,2,1 do
        if (max_spawn ==0 )or (chest.battery.energy < energy_per_action) then
          return 
        end

        max_spawn=max_spawn-1

        local line=belt[1].get_transport_line(laneI)
        local inventory = chest.inventory.get_contents()
        local itemname = next(inventory)
        
        if itemname and line~= nil and  line.can_insert_at_back() then
          local stack = {name = itemname, count = 1}
          line.insert_at_back(stack)
          chest.inventory.remove{name = itemname, count = 1}
          chest.battery.energy = chest.battery.energy - energy_per_action
        end
      end
    end
  end

end

--[[ These aren't relevant to the mod except as debugging tools.

function debugLog(message)
	if true then -- set for debug
		game.player.print(message)
	end
end

function table.val_to_str ( v )
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or
      tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{" .. table.concat( result, "," ) .. "}"
end 
--]]
